#include <string>
#include "error.h"
#include "fields.h"
#include "mathfunc.h"
#include "mathops.h"
#include "z.h"


/************/
/* field Fp */
/************/

Z Fp_prime = 2;

// operators:

// addition and substraction operators:

Fp operator+(const Fp &a) {
  return a;
}

Fp operator-(Fp a) {
  if (a != 0)
    a.value = Fp_prime - a.value;
  return a;
}

Fp operator+(const Fp &a, const Fp &b) {
  Fp temp = a;
  temp += b;
  return temp;
}

Fp operator-(const Fp &a, const Fp &b) {
  Fp temp = a;
  temp -= b;
  return temp;
}

Fp& Fp::operator++() {
  value++;
  if (value == Fp_prime)
    value = 0;
  return *this;
}

Fp Fp::operator++(int) {
  Fp temp = *this;
  ++*this;
  return temp;
}

Fp& Fp::operator--() {
  value--;
  if (value == -1)
    value += Fp_prime;
  return *this;
}

Fp Fp::operator--(int) {
  Fp temp = *this;
  --*this;
  return temp;
}

Fp Fp::operator+=(const Fp &a) {
  value += a.value;
  if (value >= Fp_prime)
    value -= Fp_prime;
  return *this;
}

Fp Fp::operator-=(const Fp &a) {
  value -= a.value;
  if (value < 0)
    value += Fp_prime;
  return *this;
}

// multiplication and division operators:

Fp operator*(const Fp &a, const Fp &b) {
  Fp temp = a;
  temp *= b;
  return temp;
}

Fp operator/(const Fp &a, const Fp &b) {
  Fp temp = a;
  temp /= b;
  return temp;
}

Fp operator%(const Fp &a, const Fp &b) {
  return 0;
}

Fp operator^(const Fp &a, const Z &b) {
  Fp temp = a;
  return temp ^= b;
}

Fp Fp::operator*=(const Fp &a) {
  value = value * a.value % Fp_prime;
  return *this;
}

Fp Fp::operator/=(const Fp &a) {
  if (a.value == 0)
    Error(DivByZero, "Fp/=Fp");
  *this *= a^(-1);
  return *this;
}

Fp Fp::operator%=(const Fp &a) {
  return 0;
}

Fp Fp::operator^=(Z a) {
  if (a < 0) {
    list<Z> L = ExtEuklid(*this, Fp_prime);
    *this = L[1];
    *this ^= -a;
  }
  else {
    Fp mult = *this, prod(1);
    for (;a != 0; a /= 2, mult *= mult)
      if (a % 2 == 1)
        prod *= mult;
    *this = prod;
  }
  return *this;
}


// comparison operators:

bool operator==(const Fp &a, const Fp &b) {
  return a.value == b.value;
}

bool operator!=(const Fp &a, const Fp &b) {
  return !(a == b);
}

bool operator<(const Fp &a, const Fp &b) {
  return a.value < b.value;
}

bool operator>(const Fp &a, const Fp &b) {
  return b < a;
}

bool operator<=(const Fp &a, const Fp &b) {
  return a < b || a == b;
}

bool operator>=(const Fp &a, const Fp &b) {
  return b < a || a == b;
}

// assignement operators:

Fp Fp::operator=(const Fp &a) {
  value = a.value;
  return *this;
}

// i/o operators:

ostream& operator<<(ostream& out, const Fp &a) {
  out << string(a);
  return out;
}

istream& operator>>(istream &in, Fp &a) {
  string s;
  in >> s;
  Fp temp(s);
  a = temp;
  return in;
}

// type conversion:

Fp::operator string() {
  return string(value);
}

Fp::operator string() const{
  return string(value);
}

Fp::operator Z() {
  return value;
}

Fp::operator Z() const{
  return value;
}

// methods:

string Fp::classtype() {
  return "field";
}

// functions:

void SetFpPrime(const Z &p) {
  if (IsPrime(p))
    Fp_prime = p;
  else
    Error(NoPrime, "SetFpPrime()");
}



/***********/
/* field Q */
/***********/

// operators:

// addition and substraction operators:

Q operator+(const Q &a) {
  return a;
}

Q operator-(Q a) {
  a.num = -a.num;
  return a;
}

Q operator+(const Q &a, const Q &b) {
  Q temp = a;
  temp += b;
  return temp;
}

Q operator-(const Q &a, const Q &b) {
  Q temp = a;
  temp -= b;
  return temp;
}

Q& Q::operator++() {
  num += den;
  return *this;
}

Q Q::operator++(int) {
  Q temp = *this;
  num += den;
  return temp;
}

Q& Q::operator--() {
  num -= den;
  return *this;
}

Q Q::operator--(int) {
  Q temp = *this;
  num -= den;
  return temp;
}

Q Q::operator+=(const Q &a) {
  Z denominator = lcm(this->den, a.den);
  num = denominator / den * num + denominator / a.den * a.num;
  den = denominator;
  reduce();
  return *this;
}

Q Q::operator-=(const Q &a) {
  return *this += -a;
}

// multiplication and division operators:

Q operator*(const Q &a, const Q &b) {
  Q temp = a;
  temp *= b;
  return temp;
}

Q operator/(const Q &a, const Q &b) {
  Q temp = a;
  temp /= b;
  return temp;
}

Q operator%(const Q &a, const Q &b) {
  return 0;
}

Q operator^(const Q &a, const Z &b) {
  Q temp = a;
  return temp ^= b;
}

Q Q::operator*=(const Q &a) {
  num *= a.num;
  den *= a.den;
  reduce();
  return *this;
}

Q Q::operator/=(const Q &a) {
  if (a.num == 0)
    Error(DivByZero, "Q/=Q");

  num *= a.den;
  den *= a.num;
  reduce();
  return *this;
}

Q Q::operator%=(const Q &a) {
  return 0;
}

Q Q::operator^=(const Z &a) {
  if (a < 0) {
    *this ^= -a;
    *this = 1 / *this;
  }
  else {
    Z temp = a;
    Q mult = *this, prod(1);
    for (; temp != 0; temp /= 2, mult *= mult)
      if (temp % 2 == 1)
        prod *= mult;
    *this = prod;
  }
  return *this;
}


// comparison operators:

bool operator==(const Q &a, const Q &b) {
  return a.num == b.num && a.den == b.den;
}

bool operator!=(const Q &a, const Q &b) {
  return !(a == b);
}

bool operator<(const Q &a, const Q &b) {
  return a.num * b.den < a.den * b.num;
}

bool operator>(const Q &a, const Q &b) {
  return b < a;
}

bool operator<=(const Q &a, const Q &b) {
  return a < b || a == b;
}

bool operator>=(const Q &a, const Q &b) {
  return b < a || a == b;
}

// assignement operators:

Q Q::operator=(const Q &a) {
  num = a.num;
  den = a.den;
  return *this;
}

// i/o operators:

ostream& operator<<(ostream& out, const Q &a) {
  out << string(a);
  return out;
}

istream& operator>>(istream &in, Q &a) {
  string s;
  in >> s;
  unsigned long l = s.length(), slashpos = 0;
  for (; s[slashpos] != '/' && slashpos < l; slashpos++);
  if (slashpos < l) {
    s[slashpos] = 0;
    Z num = s.substr(0, slashpos), den = s.substr(slashpos + 1);
    Q temp(num, den);
    a = temp;
  }
  else {
    Q temp = Z(s);
    a = temp;
  }
  return in;
}


// type conversion:

Q::operator int() {
  return int(num / den);
}

Q::operator int() const {
  return int(num / den);
}

Q::operator unsigned int() {
  return (unsigned int)(num / den);
}

Q::operator unsigned int() const {
  return (unsigned int)(num / den);
}

Q::operator long() {
  return long(num / den);
}

Q::operator long() const {
  return long(num / den);
}

Q::operator unsigned long() {
  return (unsigned long)(num / den);
}

Q::operator unsigned long() const {
  return (unsigned long)(num / den);
}

Q::operator short() {
  return short(num / den);
}

Q::operator short() const {
  return short(num / den);
}

Q::operator unsigned short() {
  return (unsigned short)(num / den);
}

Q::operator unsigned short() const {
  return (unsigned short)(num / den);
}

Q::operator string() {
  string S = string(num);
  if (den != 1)
    S = S + "/" + string(den);
  return S;
}

Q::operator string() const{
  Q temp = *this;
  return string(temp);
}

// methods:

string Q::classtype() {
  return "field";
}

void Q::reduce() {
  Z factor = gcd(num, den);
  if (factor != 1) {
    num /= factor;
    den /= factor;
  }
  if (den < 0) {
    num = -num;
    den = -den;
  }
}

