#ifndef fields_h
#define fields_h

#include <iostream>
#include <string>
#include "error.h"
#include "mathops.h"
#include "z.h"


/************/
/* field Fp */
/************/

// global prime number for all variables of type Fp
extern Z Fp_prime;

class Fp {
  friend Fp   operator- (      Fp            );
  friend bool operator==(const Fp&, const Fp&);
  friend bool operator< (const Fp&, const Fp&);

  public:

    // constructors:
    // standard constructor:
    Fp():
      value(0) {
    };
    // general constructors:
    Fp(const int &a):
      value(a % Fp_prime) {
    }
    Fp(const unsigned int &a):
      value(a % Fp_prime) {
    }
    Fp(const long &a):
      value(a % Fp_prime) {
    }
    Fp(const unsigned long &a):
      value(a % Fp_prime) {
    }
    Fp(const Z &a):
      value(a % Fp_prime) {
    }
    // copy constructor:
    Fp(const Fp &a):
      value(a.value) {
    }

    // internal operators:
    Fp& operator++(          );
    Fp  operator++(      int );
    Fp& operator--(          );
    Fp  operator--(      int );
    Fp  operator+=(const Fp &);
    Fp  operator-=(const Fp &);
    Fp  operator*=(const Fp &);
    Fp  operator/=(const Fp &);
    Fp  operator%=(const Fp &);
    Fp  operator^=(      Z   );
    Fp  operator =(const Fp &);

    // type conversion:
    operator string()      ;
    operator string() const;
    operator Z     ()      ;
    operator Z     () const;

    // methods:
    string classtype();

  private:

    Z value;

};

// external operators:

Fp       operator+ (const Fp&           );
Fp       operator- (      Fp            );
Fp       operator+ (const Fp&, const Fp&);
Fp       operator- (const Fp&, const Fp&);
Fp       operator* (const Fp&, const Fp&);
Fp       operator/ (const Fp&, const Fp&);
Fp       operator% (const Fp&, const Fp&);
Fp       operator^ (const Fp&, const Z &);
bool     operator==(const Fp&, const Fp&);
bool     operator!=(const Fp&, const Fp&);
bool     operator< (const Fp&, const Fp&);
bool     operator> (const Fp&, const Fp&);
bool     operator<=(const Fp&, const Fp&);
bool     operator>=(const Fp&, const Fp&);
ostream& operator<<(ostream &, const Fp&);
istream& operator>>(istream &,       Fp&);

// functions:
void SetFpPrime(const Z&);



/***********/
/* field Q */
/***********/

class Q {
  friend Q    operator- (      Q           );
  friend bool operator==(const Q&, const Q&);
  friend bool operator< (const Q&, const Q&);
  friend int* eval      (string            );

  public:

    // constructors:
    // standard constructor:
    Q():
      num(0),
      den(1) {
    };
    // general constructors:
    Q::Q(const int &a):
      num(a),
      den(1) {
    };
    Q::Q(const unsigned int &a):
      num(a),
      den(1) {
    };
    Q::Q(const long &a):
      num(a),
      den(1) {
    };
    Q::Q(const unsigned long &a):
      num(a),
      den(1) {
    };
    Q::Q(const Z &a):
      num(a),
      den(1) {
    };
    Q::Q(const Z &a, const Z &b):
      num(a),
      den(b) {
      if (den == 0)
        Error(DivByZero, "Q()");
      if (den < 0) {
        den = -den;
        num = -num;
      }
      reduce();
    }
    // copy constructor:
    Q::Q(const Q &a):
      num(a.num),
      den(a.den) {
    }

    // internal operators:
    Q& operator++(        );
    Q  operator++(int     );
    Q& operator--(        );
    Q  operator--(int     );
    Q  operator+=(const Q&);
    Q  operator-=(const Q&);
    Q  operator*=(const Q&);
    Q  operator/=(const Q&);
    Q  operator%=(const Q&);
    Q  operator^=(const Z&);
    Q  operator =(const Q&);

    // type conversion:
    operator            int()      ;
    operator            int() const;
    operator unsigned   int()      ;
    operator unsigned   int() const;
    operator           long()      ;
    operator           long() const;
    operator unsigned  long()      ;
    operator unsigned  long() const;
    operator          short()      ;
    operator          short() const;
    operator unsigned short()      ;
    operator unsigned short() const;
    operator string        ()      ;
    operator string        () const;

    // methods:
    string classtype();

  private:

    Z num, den; //numerator, denominator

    void reduce();
};


// external operators:

Q        operator+ (const Q&          );
Q        operator- (      Q           );
Q        operator+ (const Q&, const Q&);
Q        operator- (const Q&, const Q&);
Q        operator* (const Q&, const Q&);
Q        operator/ (const Q&, const Q&);
Q        operator% (const Q&, const Q&);
Q        operator^ (const Q&, const Z&);
bool     operator==(const Q&, const Q&);
bool     operator!=(const Q&, const Q&);
bool     operator< (const Q&, const Q&);
bool     operator> (const Q&, const Q&);
bool     operator<=(const Q&, const Q&);
bool     operator>=(const Q&, const Q&);
ostream& operator<<(ostream&, const Q&);
istream& operator>>(istream&,       Q&);


#endif

