#include <dos.h>
#include <stdlib.h>
#include <string>
#include "funcs.h"
#include "z.h"

// string functions:

unsigned int counttokens(const string s, const char &c) {
  unsigned int i = 0, l = s.length(), count = 0;
  string token = "";
  for (; i < l; i++)
    if (s[i] == c) {
      if (token != "") {
        count++;
        token = "";
      }
    }
    else
      token += s[i];
  if (token != "")
    count++;
  return count;
}

string gettoken(const string s, const char &c, const unsigned int &pos) {
  unsigned int i = 0, l = s.length(), count = 1;
  string token = "";
  for (; i < l; i++)
    if (s[i] == c) {
      if (token != "") {
        if (count == pos)
          return token;
        count++;
        token = "";
      }
    }
    else
      token += s[i];
  if (token != "" && count == pos)
    return token;
  return "";
}

bool isin(const string &s, const char &c) {
  unsigned int i = 0, l = s.length();
  for (; i < l; i++)
    if (s[i] == c)
      return true;
  return false;
}

bool isletter(const char &c) {
  return c >= 'A' && c <= 'Z' || c >= 'a' && c <= 'z';
}

bool isnumeric(const string &s) {
  unsigned int i = 0, l = s.length();
  if (l == 0)
    return false;
  if (s[0] == '+' || s[0] == '-')
    for (i = 1; i < l && s[i] == ' '; i++);
  for (; i < l; i++)
    if (!isdigit(s[i]))
      return false;
  return true;
}

string left(const string &s, const unsigned int &len) {
  return s.substr(0, len);
}

string lower(string s) {
  unsigned int i = 0, l = s.length();
  for (; i < l; i++)
    if (s[i] > 64 && s[i] < 91)
      s[i] += 32;
  return s;
}

string ltrim(const string s) {
  unsigned int i = 0, l = s.length();
  while (i < l && (s[i] == '\t' || s[i] == ' '))
    i++;
  return s.substr(i);
}

string replicate(const char &c, const unsigned int &n) {
  string s;
  for (unsigned int i = 1; i <= n; i++)
    s += c;
  return s;
}

string repsubstring(string s, const string &oldstring, const string &newstring) {
  unsigned int i = 0, ln = newstring.length(), lo = oldstring.length();
  string right;
  while ((i = s.find(oldstring, i)) != -1u) {
    right = s.substr(lo + i);
    s     = s.substr(0,   i);
    s    += newstring;
    s    += right    ;
    i    += ln       ;
  }
  return s;
}

string right(const string &s, const unsigned int &len) {
  unsigned int l = s.length();
  if (l <= len)
    return s;
  else
    return s.substr(l - len);
}

string rtrim(const string s) {
  unsigned int i = s.length() - 1;
  while (i != -1u && (s[i] == '\t' || s[i] == ' '))
    i--;
  return s.substr(0, i + 1);
}

string space(const unsigned int &n) {
  string s;
  for (unsigned int i = 1; i <= n; i++)
    s += ' ';
  return s;
}

string str(Z a, const unsigned long &l) {
  string pre, s;

  if (a == 0)
    s = "0";
  else {
    if (a < 0) {
      pre = '-';
      a   = -a ;
    }
    while (a != 0) {
      s = char('0' + int(a % 10)) + s;
      a /= 10;
    }
    s = pre + s;
  }

  if (l > 0)
    if (l < s.length())
      s = right(s, l);
    else
      while (l > s.length())
        s = '0' + s;

  return s;
}

string trim(const string s) {
  return ltrim(rtrim(s));
}

string upper(string s) {
  unsigned int i = 0, l = s.length();
  for (; i < l; i++)
    if (s[i] > 96 && s[i] < 123)
      s[i] -= 32;
  return s;
}

Z val(const string &s) {
  unsigned int last, l = s.length();
  Z result = 0, factor;
  char c;
  if (l == 0)
    return 0;
  l--;
  char prefix = s[0];
  switch (prefix) {
    case '+': last   = 0;
              factor = 1;
              break;
    case '-': last   =  0;
              factor = -1;
              break;
    default:  last   = -1u;
              factor =  1 ;
  }
  for (; last < l && s[last + 1] == ' '; last++);
  for (; l != last; l--, factor *= 10) {
    c = s[l];
    if (c < '0' || c > '9')
      return 0;
    else
      result += factor * int(c - '0');
  }
  return result;
}

// other functions:

unsigned int CurrentDate() {
  date today;
  unsigned int d = 0;
  getdate(&today);
  d += today.da_year;
  d *= 100          ;
  d += today.da_mon ;
  d *= 100          ;
  d += today.da_day ;
  return d;
}

unsigned int CurrentTime() {
  time now;
  unsigned int t = 0;
  gettime(&now);
  t += now.ti_hour;
  t *= 60         ;
  t += now.ti_min ;
  t *= 60         ;
  t += now.ti_sec ;
  t *= 100        ;
  t += now.ti_hund;
  return t;
}

string FormatTime(int t) {
  string Time;
  int h, m, s, x;

  x  = t % 100;
  t /=     100;
  s  = t %  60;
  t /=      60;
  m  = t %  60;
  h  = t /  60;
  if (h > 0) {
    Time += str(h);
    Time += "h "  ;
  }
  if (m > 0) {
    Time += str(m);
    Time += "m "  ;
  }
  Time += str(s)   ;
  Time += "."      ;
  Time += str(x, 2);
  Time += "s"      ;
  return Time;
}

Z Rand(const Z &a, const Z &b) {
  if (a == b)
    return a;
  if (a > b)
    return Rand(b, a);

  Z m = b - a, r;
  unsigned long i, l = string(m).length();
  Z M = (Z(10)^l) / m * m;
  while (true) {
    r = 0;
    for (i = 1; i <= l; i++)
      r = 10 * r + random() % 10;
    if (r <= M)
      return a + r % m;
  }
}

void RandSeed() {
  srandom(CurrentTime());
  srandom(random     ());
}

int timer(const char &command) {
  static int lastdate;
  static int lasttime;
  int diff  = 0            ,
      now   = CurrentTime(),
      today = CurrentDate();

  switch(command) {
    case 'e': diff = now - lasttime;
              if (today != lastdate)
                diff += 8640000;
    case 's': lastdate = today;
              lasttime = now  ;
  }
  return diff;
}

