#include <fstream>
#include <iostream>
#include <string>
#include "contain.t"
#include "fields.h"
#include "funcs.h"
#include "gb.t"
#include "ideal.t"
#include "mathfunc.h"
#include "mathops.h"
#include "parser.h"
#include "poly.h"
#include "z.h"

bool app = false, debug = false, debug2 = false, dooutput = false;

bool          Homo_Error         ,
              Parser_Quit = false,
              Show_Stats  = false;
char          Eval_Type;
istream       din  = cin ;
ostream       dout = cerr;
unsigned long Exp_Depth = 0;

// set a parser variable

bool assignvariable(const string &var, string &expr) {
  string upvar = upper(var);

  if (trim(var) == "") {
    cout << "Please give a variable name!" << endl << endl;
    help("define");
    return false;
  }
  if (var.length() == 1 && IndetList.search(var[0])) {
    cout << '"' << var << "\" is an indeterminate and therefore cannot be used as a variable name!" << endl;
    return false;
  }
  if (upvar == "AND" || upvar == "OR" || upvar == "NOT") {
    cout << '"' << var << "\" is a reserved word and therefore cannot be used as a variable name!" << endl;
    return false;
  }
  if (expr == "") {
    cout << "Please give an expression to assign to variable \"" << var << "\"!" << endl << endl;
    help("define");
    return false;
  }

  expr = repsubstring(trim(expr), "\t", " ");
  int* value = eval(expr);
  if (Eval_Type == '-') {
    cout << "\"" << expr << "\" is not a valid expression!" << endl;
    return false;
  }

  for (unsigned long i = 1; i <= variables.length(); i++)
    if (var == variables[i].name) {
      switch (variables[i].type) {
        case 'B': delete (bool*)variables[i].data;
                  break;
        case 'N': switch (Field) {
                    case RAT: delete (Q*)variables[i].data;
                              break;
                    case MOD: delete (Fp*)variables[i].data;
                              break;
                  }
                  break;
        case 'P': switch (Field) {
                    case RAT: delete (poly<Q>*)variables[i].data;
                              break;
                    case MOD: delete (poly<Fp>*)variables[i].data;
                              break;
                  }
                  break;
        case 'I': switch (Field) {
                    case RAT: delete (ideal<Q>*)variables[i].data;
                              break;
                    case MOD: delete (ideal<Fp>*)variables[i].data;
                              break;
                  }
                  break;
        case 'G': switch (Field) {
                    case RAT: delete (list<Q>*)variables[i].data;
                              break;
                    case MOD: delete (list<Fp>*)variables[i].data;
                              break;
                  }
                  break;
      }
      variables[i] = variable(var, Eval_Type, value);
      return true;
    }
  variables.append(variable(var, Eval_Type, value));
  return true;
}


// deleting a parser variable from memory

bool deletevariable(const string &var) {
  if (trim(var) == "") {
    cout << "Please give the name of the user variable you want to remove from memory!" << endl << endl;
    help("undefine");
    return false;
  }
  for (unsigned long i = 1; i <= variables.length(); i++)
    if (var == variables[i].name) {
      switch (variables[i].type) {
        case 'B': delete (bool*)variables[i].data;
                  break;
        case 'N': switch (Field) {
                    case RAT: delete (Q*)variables[i].data;
                              break;
                    case MOD: delete (Fp*)variables[i].data;
                              break;
                  }
                  break;
        case 'P': switch (Field) {
                    case RAT: delete (poly<Q>*)variables[i].data;
                              break;
                    case MOD: delete (poly<Fp>*)variables[i].data;
                              break;
                  }
                  break;
        case 'I': switch (Field) {
                    case RAT: delete (ideal<Q>*)variables[i].data;
                              break;
                    case MOD: delete (ideal<Fp>*)variables[i].data;
                              break;
                  }
                  break;
        case 'G': switch (Field) {
                    case RAT: delete (list<Q>*)variables[i].data;
                              break;
                    case MOD: delete (list<Fp>*)variables[i].data;
                              break;
                  }
                  break;
      }
      variables.remove(i);
      return true;
    }
  cout << "There was no user variable \"" << var << "\" found in memory!" << endl;
  return false;
}

// display expression

bool display(const string &expr) {
  int              *p = eval(expr);
  bool             *pb ;
  Q                *pnq;
  Fp               *pnf;
  poly <     Q   > *ppq;
  poly <     Fp  > *ppf;
  ideal<     Q   > *piq;
  ideal<     Fp  > *pif;
  list <poly<Q > > *pgq;
  list <poly<Fp> > *pgf;
  if (trim(expr) == "") {
    cout << "Please give an expression to display!" << endl << endl;
    help("print");
    return false;
  }
  if (Eval_Type == '-') {
    cout << "\"" << expr << "\" is not a valid expression!" << endl;
    return false;
  }

  switch (Eval_Type) {
    case 'B': pb = (bool*)p;
              cout << (*pb ? "TRUE" : "FALSE") << endl;
              delete pb;
              return true;
    case 'N': switch (Field) {
                case RAT: pnq = (Q*)p;
                          cout << *pnq << endl;
                          delete pnq;
                          return true;
                case MOD: pnf = (Fp*)p;
                          cout << *pnf << endl;
                          delete pnf;
                          return true;
              }
    case 'P': switch (Field) {
                case RAT: ppq = (poly<Q>*)p;
                          cout << *ppq << endl;
                          delete ppq;
                          return true;
                case MOD: ppf = (poly<Fp>*)p;
                          cout << *ppf << endl;
                          delete ppf;
                          return true;
              }
    case 'I': switch (Field) {
                case RAT: piq = (ideal<Q>*)p;
                          cout << *piq << endl;
                          delete piq;
                          return true;
                case MOD: pif = (ideal<Fp>*)p;
                          cout << *pif << endl;
                          delete pif;
                          return true;
              }
    case 'G': switch (Field) {
                case RAT: pgq = (list<poly<Q> >*)p;
                          cout << *pgq << endl;
                          delete pgq;
                          return true;
                case MOD: pgf = (list<poly<Fp> >*)p;
                          cout << *pgf << endl;
                          delete pgf;
                          return true;
              }
  }
  return false;
}


// evaluate expression

int* eval(string expr) {
  int              *p  , *q  ;
  bool             *pb , *qb , acceptable;
  Q                *pnq, *qnq, PNQ, QNQ;
  Fp               *pnf, *qnf, PNF, QNF;
  poly <     Q   > *ppq, *qpq, PPQ, QPQ;
  poly <     Fp  > *ppf, *qpf, PPF, QPF;
  ideal<     Q   > *piq, *qiq          ;
  ideal<     Fp  > *pif, *qif          ;
  list <poly<Q > > *pgq, *qgq, PLQ     ;
  list <poly<Fp> > *pgf, *qgf, PLF     ;
  unsigned int pos, BrackDepth = 0, l, n;
  unsigned long i, j;
  string bracket, brackets, exponent, idealexpr, indet, listexpr, operation,
         polyexpr, subexpr, symbols, typestring, upexpr;

  expr   = trim (expr);
  upexpr = upper(expr);
  if (expr[0] == '(' && expr[expr.length() - 1] == ')')
    expr = expr.substr(1, expr.length() - 2);
  symbols = expr;
  l = symbols.length();
  bracket = "";
  for (i = 0; i < l; i++) {
    if (bracket == "" || bracket == "(")
      switch (symbols[i]) {
        case '(': BrackDepth++;
                  bracket = "(";
                  if (BrackDepth > 1)
                    symbols[i] = 'P';
                  break;
        case ')': BrackDepth--;
                  switch (BrackDepth) {
                    case -1u: Eval_Type = '-';
                              return NULL;
                    case  0 : bracket = "";
                  }
        default : if (BrackDepth > 0)
                    symbols[i] = 'P';
      }
    if (bracket == "" || bracket == "[")
      switch (symbols[i]) {
        case '[': BrackDepth++;
                  bracket = "[";
                  if (BrackDepth > 1)
                    symbols[i] = 'B';
                  break;
        case ']': BrackDepth--;
                  switch (BrackDepth) {
                    case -1u: Eval_Type = '-';
                              return NULL;
                    case  0 : bracket = "";
                  }
        default : if (BrackDepth > 0)
                    symbols[i] = 'B';
      }
  }

  // variable?
  pos = evalfindstring(expr, "[");
  subexpr = pos == -1u ? expr : left(expr, pos);
  for (i = 1; i <= variables.length(); i++)
    if (subexpr == variables[i].name) {
      Eval_Type = variables[i].type;
      if (expr == subexpr || Eval_Type == 'G')
        switch (Eval_Type) {
          case 'B': pb  = new bool;
                    *pb = *(bool*)variables[i].data;
                    return (int*)pb;
          case 'N': switch (Field) {
                      case RAT: pnq  = new Q;
                                *pnq = *(Q*)variables[i].data;
                                return (int*)pnq;
                      case MOD: pnf  = new Fp;
                                *pnf = *(Fp*)variables[i].data;
                                return (int*)pnf;
                    }
          case 'P': switch (Field) {
                      case RAT: ppq  = new poly<Q>;
                                *ppq = *(poly<Q>*)variables[i].data;
                                return (int*)ppq;
                      case MOD: ppf  = new poly<Fp>;
                                *ppf = *(poly<Fp>*)variables[i].data;
                                return (int*)ppf;
                    }
          case 'I': switch (Field) {
                      case RAT: piq  = new ideal<Q>;
                                *piq = *(ideal<Q>*)variables[i].data;
                                return (int*)piq;
                      case MOD: pif  = new ideal<Fp>;
                                *pif = *(ideal<Fp>*)variables[i].data;
                                return (int*)pif;
                    }
          case 'G': if (expr == subexpr)
                      switch (Field) {
                        case RAT: pgq  = new list<poly<Q> >;
                                  *pgq = *(list<poly<Q> >*)variables[i].data;
                                  return (int*)pgq;
                        case MOD: pgf  = new list<poly<Fp> >;
                                  *pgf = *(list<poly<Fp> >*)variables[i].data;
                                  return (int*)pgf;
                      }
                    else {
                      subexpr = expr.substr(pos);
                      subexpr = getbrackets(subexpr, "[");
                      if (subexpr.length() != expr.length() - pos - 2)
                        break;
                      p = eval(subexpr);
                      if (Eval_Type == 'N') {
                        pnq = (Q*)p;
                        if ((*pnq).den == 1)
                          bracket = str((*pnq).num);
                        else
                          bracket = "";
                        delete pnq;
                        if (bracket == "") {
                          release(p, Eval_Type);
                          break;
                        }
                        else {
                          j = val(bracket);
                          switch (Field) {
                            case RAT: if (j != 0 && j <= (*(list<poly<Q> >*)variables[i].data).length()) {
                                        Eval_Type = 'P';
                                        ppq  = new poly<Q>;
                                        *ppq = (*(list<poly<Q> >*)variables[i].data)[j];
                                        return (int*)ppq;
                                      }
                                      break;
                            case MOD: if (j != 0 && j <= (*(list<poly<Fp> >*)variables[i].data).length()) {
                                        Eval_Type = 'P';
                                        ppf  = new poly<Fp>;
                                        *ppf = (*(list<poly<Fp> >*)variables[i].data)[j];
                                        return (int*)ppf;
                                      }
                                      break;
                          }
                          release(p, Eval_Type);
                          break;
                        }
                      }
                      else {
                        release(p, Eval_Type);
                        break;
                      }
                    }
        }
    }

  // function?
  if (left(upexpr, 2) == "GB") {
    GB_Type gbt = GBType;
    subexpr = trim(expr.substr(2));
    subexpr = getbrackets(subexpr, "(");
    if (counttokens(subexpr, ',') > 1 && right(subexpr, 1) != ")") {
      typestring = lower(trim(gettoken(subexpr, ',', counttokens(subexpr, ','))));
      subexpr    = left(subexpr, subexpr.length() - typestring.length());
      if (typestring == "b" || typestring == "buch" || typestring == "buchberger")
        gbt = Buchberger;
      else if (typestring == "bh" || typestring == "buchhomo" || typestring == "buchbergerhomo")
        gbt = BuchbergerHomo;
      else if (typestring == "bg" || typestring == "buchgauss" || typestring == "buchbergergauss")
        gbt = BuchbergerGauss;
      else if (typestring == "bm" || typestring == "buchmulti" || typestring == "buchbergermulti")
        gbt = BuchbergerMulti;
      else if (typestring == "f4" || typestring == "fau" || typestring == "faugere")
        gbt = F4;
    }
    if (subexpr != "") {
      p = eval(subexpr);
      if (Eval_Type == 'I') {
        switch (Field) {
          case RAT: piq  = (ideal<Q>*)p;
                    pgq  = new list<poly<Q> >;
                    *pgq = GB(*piq, gbt);
                    delete piq;
                    if ((gbt == BuchbergerHomo || gbt == BuchbergerGauss || gbt == BuchbergerMulti || gbt == F4) && Homo_Error) {
                      cout << "This ideal seems not to be homogeneous!" << endl;
                      Eval_Type = '-';
                      return NULL;
                    }
                    Eval_Type = 'G';
                    return (int*)pgq;
                    break;
          case MOD: pif  = (ideal<Fp>*)p;
                    pgf  = new list<poly<Fp> >;
                    *pgf = GB(*pif, gbt);
                    delete pif;
                    if ((gbt == BuchbergerHomo || gbt == BuchbergerGauss || gbt == BuchbergerMulti || gbt == F4) && Homo_Error) {
                      cout << "This ideal seems not to be homogeneous!" << endl;
                      Eval_Type = '-';
                      return NULL;
                    }
                    Eval_Type = 'G';
                    return (int*)pgf;
                    break;
        }
      }
      else
        release(p, Eval_Type);
    }
  }
  if (left(upexpr, 3) == "LEN") {
    subexpr = trim(expr.substr(3));
    subexpr = getbrackets(subexpr, "(");
    if (subexpr != "") {
      p = eval(subexpr);
      if (Eval_Type == 'G') {
        pnq = new Q;
        switch (Field) {
          case RAT: pgq  = (list<poly<Q> >*)p;
                    *pnq = pgq->length();
                    delete pgq;
                    break;
          case MOD: pgf  = (list<poly<Fp> >*)p;
                    *pnq = pgf->length();
                    delete pgf;
                    break;
        }
        Eval_Type = 'N';
        return (int*)pnq;
      }
      else
        release(p, Eval_Type);
    }
  }

  // bool?
  if (upexpr == "TRUE") {
    pb  = new bool;
    *pb = true;
    Eval_Type = 'B';
    return (int*)pb;
  }
  if (upexpr == "FALSE") {
    pb  = new bool;
    *pb = false;
    Eval_Type = 'B';
    return (int*)pb;
  }

  // number?
  if (isnumeric(expr))
    if (Field == RAT || Exp_Depth > 0) {
      pnq  = new Q;
      *pnq = val(expr);
      Eval_Type = 'N';
      return (int*)pnq;
    }
    else if (Field == MOD) {
      pnf  = new Fp;
      *pnf = val(expr);
      Eval_Type = 'N';
      return (int*)pnf;
    }

  // polynomial?
  if (left(upexpr, 4) == "POLY") {
    polyexpr = expr.substr(4);
    polyexpr = getbrackets(polyexpr, "(");
    p = eval(polyexpr);
    if (Eval_Type == 'N') {
      switch (Field) {
        case RAT: pnq  = (Q*)p;
                  ppq  = new poly<Q>;
                  *ppq = *pnq;
                  delete pnq;
                  Eval_Type = 'P';
                  return (int*)ppq;
        case MOD: pnf  = (Fp*)p;
                  ppf  = new poly<Fp>;
                  *ppf = *pnf;
                  delete pnf;
                  Eval_Type = 'P';
                  return (int*)ppf;
      }
    }
    else
      release(p, Eval_Type);
  }
  else {
    polyexpr = expr;
    subexpr  = getoperator(polyexpr);
    subexpr += getnumber(polyexpr);
    switch (Field) {
      case RAT: if (subexpr == "")
                  PNQ = 1;
                else if (subexpr == "-")
                  PNQ = -1;
                else {
                  PNQ = val(subexpr);
                  if (left(polyexpr, 1) == "/") {
                    polyexpr = polyexpr.substr(1);
                    subexpr  = getnumber(polyexpr);
                    QNQ = val(subexpr);
                    if (QNQ == 0) {
                      polyexpr = "";
                      break;
                    }
                    else
                      PNQ /= QNQ;
                  }
                }
                break;
      case MOD: if (subexpr == "")
                  PNF = 1;
                else if (subexpr == "-")
                  PNF = -1;
                else {
                  PNF = val(subexpr);
                  if (left(polyexpr, 1) == "/") {
                    polyexpr = polyexpr.substr(1);
                    subexpr  = getnumber(polyexpr);
                    QNF = val(subexpr);
                    if (QNF == 0) {
                      polyexpr = "";
                      break;
                    }
                    else
                      PNF /= QNF;
                  }
                }
                break;
    }
    acceptable = polyexpr.length() > 0 && left(polyexpr, 1) != "[";
    if (acceptable)
      for (i = 0; i < polyexpr.length(); i++) {
        switch (polyexpr[i]) {
          case '[':
                    subexpr  = left(polyexpr, i);
                    polyexpr = polyexpr.substr(i);
                    bracket  = getbrackets(polyexpr, "[");
                    brackets = "[";
                    brackets += bracket;
                    brackets += "]";
                    Exp_Depth++;
                    p = eval(bracket);
                    Exp_Depth--;
                    bracket = "";
                    if (Eval_Type == 'N') {
                      pnq = (Q*)p;
                      if ((*pnq).den == 1) {
                        bracket = "[";
                        bracket += str((*pnq).num);
                        bracket += "]";
                      }
                      delete pnq;
                    }
                    else
                      release(p, Eval_Type);
                    if (bracket == "")
                      acceptable = false;
                    else {
                      subexpr += repsubstring(polyexpr, brackets, bracket);
                      polyexpr = subexpr;
                    }
                    break;
          case '^': subexpr  = left(polyexpr, i);
                    polyexpr = polyexpr.substr(i + 1);
                    exponent = getbrackets(polyexpr, "(");
                    if (exponent == "")
                      exponent = getnumber(polyexpr);
                    Exp_Depth++;
                    p = eval(exponent);
                    Exp_Depth--;
                    exponent = "";
                    if (Eval_Type == 'N') {
                      pnq = (Q*)p;
                      if ((*pnq).den == 1)
                        exponent = str((*pnq).num);
                      delete pnq;
                    }
                    else
                      release(p, Eval_Type);
                    if (exponent == "")
                      acceptable = false;
                    else {
                      subexpr += '^';
                      subexpr += exponent;
                      subexpr += polyexpr;
                      polyexpr = subexpr;
                    }
                    break;
        }
        if (!acceptable)
          break;
      }
    if (acceptable) {
      subexpr = "";
      while (polyexpr.length() != 0) {
        indet = getindet(polyexpr);
        if (indet == "") {
          acceptable = false;
          break;
        }
        subexpr += indet;
        if (polyexpr.length() > 0 && polyexpr[0] == '^') {
          polyexpr = polyexpr.substr(1);
          exponent = getnumber(polyexpr);
          if (exponent == "") {
            acceptable = false;
            break;
          }
          subexpr += '^';
          subexpr += exponent;
        }
      }
      if (acceptable)
        switch (Field) {
          case RAT: ppq  = new poly<Q>;
                    *ppq = term<Q>(PNQ, monomial(subexpr));
                    Eval_Type = 'P';
                    return (int*)ppq;
          case MOD: ppf  = new poly<Fp>;
                    *ppf = term<Fp>(PNF, monomial(subexpr));
                    Eval_Type = 'P';
                    return (int*)ppf;
        }
    }
  }

  // ideal?
  if (left(upexpr, 5) == "IDEAL") {
    idealexpr = trim(expr.substr(5));
    if (left(idealexpr, 1) == "(" && right(idealexpr, 1) == ")") {
      idealexpr = idealexpr.substr(1, idealexpr.length() - 2);
      if (left(idealexpr, 1) != "[" && counttokens(idealexpr, ',') > 1)
        idealexpr = '[' + idealexpr + ']';
      p = eval(idealexpr);
      if (Eval_Type == 'G')
        switch (Field) {
          case RAT: pgq  = (list<poly<Q> >*)p;
                    piq  = new ideal<Q>;
                    *piq = ideal<Q>(*pgq);
                    delete pgq;
                    Eval_Type = 'I';
                    return (int*)piq;
          case MOD: pgf  = (list<poly<Fp> >*)p;
                    pif  = new ideal<Fp>;
                    *pif = ideal<Fp>(*pgf);
                    delete pgf;
                    Eval_Type = 'I';
                    return (int*)pif;
        }
      release(p, Eval_Type);
      n = counttokens(idealexpr, ',');
      acceptable = true;
      switch (Field) {
        case RAT: PLQ = list<poly<Q> >(n, poly<Q>(0));
                  for (i = 1; i <= n; i++) {
                    p = eval(gettoken(idealexpr, ',', i));
                    switch (Eval_Type) {
                      case 'N': pnq  = new Q;
                                pnq  = (Q*)p;
                                ppq  = new poly<Q>;
                                *ppq = poly<Q>(*pnq);
                                delete pnq;
                                break;
                      case 'P': ppq = new poly<Q>;
                                ppq = (poly<Q>*)p;
                                break;
                      default:  acceptable = false;
                                release(p, Eval_Type);
                    }
                    if (acceptable) {
                      PLQ[i] = *ppq;
                      delete ppq;
                    }
                    else
                      break;
                  }
                  if (acceptable) {
                    piq  = new ideal<Q>;
                    *piq = ideal<Q>(PLQ);
                    Eval_Type = 'I';
                    return (int*)piq;
                  }
                  break;
        case MOD: PLF = list<poly<Fp> >(n, poly<Fp>(0));
                  for (i = 1; i <= n; i++) {
                    p = eval(gettoken(idealexpr, ',', i));
                    switch (Eval_Type) {
                      case 'N': pnf  = new Fp;
                                pnf  = (Fp*)p;
                                ppf  = new poly<Fp>;
                                *ppf = poly<Fp>(*pnf);
                                delete pnf;
                                break;
                      case 'P': ppf = new poly<Fp>;
                                ppf = (poly<Fp>*)p;
                                break;
                      default:  acceptable = false;
                                release(p, Eval_Type);
                    }
                    if (acceptable) {
                      PLF[i] = *ppf;
                      delete ppf;
                    }
                    else
                      break;
                  }
                  if (acceptable) {
                    pif  = new ideal<Fp>;
                    *pif = ideal<Fp>(PLF);
                    Eval_Type = 'I';
                    return (int*)pif;
                  }
                  break;
      }
    }
  }

  // list?
  if (left(expr, 1) == "[" && right(expr, 1) == "]" && repsubstring(symbols, "B", "") == "[]") {
    listexpr = expr.substr(1, expr.length() - 2);
    n = counttokens(listexpr, ',');
    acceptable = true;
    switch (Field) {
      case RAT: PLQ = list<poly<Q> >(n, poly<Q>(0));
                for (i = 1; i <= n; i++) {
                  subexpr = gettoken(listexpr, ',', i);
                  p = eval(subexpr);
                  switch (Eval_Type) {
                    case 'N': pnq  = new Q;
                              pnq  = (Q*)p;
                              ppq  = new poly<Q>;
                              *ppq = poly<Q>(*pnq);
                              delete pnq;
                              break;
                    case 'P': ppq = new poly<Q>;
                              ppq = (poly<Q>*)p;
                              break;
                    default:  acceptable = false;
                              release(p, Eval_Type);
                  }
                  if (acceptable) {
                    PLQ[i] = *ppq;
                    delete ppq;
                  }
                  else
                    break;
                }
                if (acceptable) {
                  pgq  = new list<poly<Q> >;
                  *pgq = PLQ;
                  Eval_Type = 'G';
                  return (int*)pgq;
                }
                break;
      case MOD: PLF = list<poly<Fp> >(n, poly<Fp>(0));
                for (i = 1; i <= n; i++) {
                  p = eval(gettoken(listexpr, ',', i));
                  switch (Eval_Type) {
                    case 'N': pnf  = new Fp;
                              pnf  = (Fp*)p;
                              ppf  = new poly<Fp>;
                              *ppf = poly<Fp>(*pnf);
                              delete pnf;
                              break;
                    case 'P': ppf = new poly<Fp>;
                              ppf = (poly<Fp>*)p;
                              break;
                    default:  acceptable = false;
                              release(p, Eval_Type);
                  }
                  if (acceptable) {
                    PLF[i] = *ppf;
                    delete ppf;
                  }
                  else
                    break;
                }
                if (acceptable) {
                  pgf  = new list<poly<Fp> >;
                  *pgf = PLF;
                  Eval_Type = 'G';
                  return (int*)pgf;
                }
                break;
    }
  }

  // search for "or"
  pos = evalfindstring(upper(symbols), " OR ");
  if (pos != -1u) {
    p = eval(expr.substr(0, pos));
    if (Eval_Type != 'B') {
      release(p, Eval_Type);
      Eval_Type = '-';
      return NULL;
    }
    q = eval(expr.substr(pos + 4));
    if (Eval_Type != 'B') {
      release(p, 'B'      );
      release(q, Eval_Type);
      Eval_Type = '-';
      return NULL;
    }
    pb = (bool*)p;
    qb = (bool*)q;
    *pb = *pb || *qb;
    delete qb;
    return p;
  }

  // search for "and"
  pos = evalfindstring(upper(symbols), " AND ");
  if (pos != -1u) {
    p = eval(expr.substr(0, pos ));
    if (Eval_Type != 'B') {
      release(p, Eval_Type);
      Eval_Type = '-';
      return NULL;
    }
    q = eval(expr.substr(pos + 5));
    if (Eval_Type != 'B') {
      release(p, 'B'      );
      release(q, Eval_Type);
      Eval_Type = '-';
      return NULL;
    }
    pb = (bool*)p;
    qb = (bool*)q;
    *pb = *pb && *qb;
    delete qb;
    return p;
  }

  // search for "not"
  pos = evalfindstring(upper(symbols), "NOT ");
  if (pos != -1u && (pos == 0 || !isletter(symbols[pos - 1]) && symbols[pos - 1] != '_')) {
    q = eval(expr.substr(pos + 4));
    if (Eval_Type != 'B') {
      release(q, Eval_Type);
      Eval_Type = '-';
      return NULL;
    }
    qb = (bool*)q;
    *qb = !*qb;
    return q;
  }

  // search for "=", "<>", "!="
  pos = evalfindstring(symbols, "!=");
  if (pos == -1u) {
    pos = evalfindstring(symbols, "<>");
    if (pos == -1u) {
      pos = evalfindstring(symbols, "=");
      if (pos != -1u && (pos == 0 || expr[pos - 1] != '<' && expr[pos - 1] != '>'))
        operation = "=";
      else
        pos = -1u;
    }
    else
      operation = "<>";
  }
  else
    operation = "<>";
  if (pos != -1u) {
    p = eval(expr.substr(0, pos));
    typestring = Eval_Type;
    q = eval(expr.substr(pos + operation.length()));
    typestring += Eval_Type;
    if (typestring == "BB") {
      pb = (bool*)p;
      qb = (bool*)q;
      *pb = (operation == "=" ? *pb == *qb : *pb != *qb);
      delete qb;
      return p;
    }
    else if (typestring == "NN") {
      pb = new bool;
      switch (Field) {
        case RAT: pnq = (Q*)p;
                  qnq = (Q*)q;
                  *pb = (operation == "=" ? *pnq == *qnq : *pnq != *qnq);
                  delete pnq, qnq;
                  Eval_Type = 'B';
                  break;
        case MOD: pnf = (Fp*)p;
                  qnf = (Fp*)q;
                  *pb = (operation == "=" ? *pnf == *qnf : *pnf != *qnf);
                  delete pnf, qnf;
                  Eval_Type = 'B';
                  break;
      }
      return (int*)pb;
    }
    else if (typestring == "NP") {
      pb = new bool;
      switch (Field) {
        case RAT: pnq = (     Q *)p;
                  qpq = (poly<Q>*)q;
                  *pb = (operation == "=" ? poly<Q>(*pnq) == *qpq : poly<Q>(*pnq) != *qpq);
                  delete pnq, qpq;
                  Eval_Type = 'B';
                  break;
        case MOD: pnf = (     Fp *)p;
                  qpf = (poly<Fp>*)q;
                  *pb = (operation == "=" ? poly<Fp>(*pnf) == *qpf : poly<Fp>(*pnf) != *qpf);
                  delete pnf, qpf;
                  Eval_Type = 'B';
                  break;
      }
      return (int*)pb;
    }
    else if (typestring == "PN") {
      pb = new bool;
      switch (Field) {
        case RAT: ppq = (poly<Q>*)p;
                  qnq = (     Q *)q;
                  *pb = (operation == "=" ? *ppq == poly<Q>(*qnq) : *ppq != poly<Q>(*qnq));
                  delete ppq, qnq;
                  Eval_Type = 'B';
                  break;
        case MOD: ppf = (poly<Fp>*)p;
                  qnf = (     Fp *)q;
                  *pb = (operation == "=" ? *ppf == poly<Fp>(*qnf) : *ppf != poly<Fp>(*qnf));
                  delete ppf, qnf;
                  Eval_Type = 'B';
                  break;
      }
      return (int*)pb;
    }
    else if (typestring == "PP") {
      pb = new bool;
      switch (Field) {
        case RAT: ppq = (poly<Q>*)p;
                  qpq = (poly<Q>*)q;
                  *pb = (operation == "=" ? *ppq == *qpq : *ppq != *qpq);
                  delete ppq, qpq;
                  Eval_Type = 'B';
                  break;
        case MOD: ppf = (poly<Fp>*)p;
                  qpf = (poly<Fp>*)q;
                  *pb = (operation == "=" ? *ppf == *qpf : *ppf != *qpf);
                  delete ppf, qpf;
                  Eval_Type = 'B';
                  break;
      }
      return (int*)pb;
    }
    //else if (typestring == "II") {
    //}
    else if (typestring == "GG") {
      pb = new bool;
      switch (Field) {
        case RAT: pgq = (list<poly<Q> >*)p;
                  qgq = (list<poly<Q> >*)q;
                  *pb = (operation == "=" ? *pgq == *qgq : *pgq != *qgq);
                  delete pgq, qgq;
                  Eval_Type = 'B';
                  break;
        case MOD: pgf = (list<poly<Fp> >*)p;
                  qgf = (list<poly<Fp> >*)q;
                  *pb = (operation == "=" ? *pgf == *qgf : *pgf != *qgf);
                  delete pgf, qgf;
                  Eval_Type = 'B';
                  break;
      }
      return (int*)pb;
    }
    else {
      release(p, typestring[0]);
      release(q, typestring[1]);
      Eval_Type = '-';
      return NULL;
    }
  }

  // search for "<", ">", "<=", ">="
  pos = evalfindstring(symbols, "<=");
  if (pos == -1u) {
    pos = evalfindstring(symbols, "<");
    if (pos == -1u) {
      pos = evalfindstring(symbols, ">=");
      if (pos == -1u) {
        pos = evalfindstring(symbols, ">");
        if (pos != -1u)
          operation = ">";
      }
      else
        operation = ">=";
    }
    else
      operation = "<";
  }
  else
    operation = "<=";
  if (pos != -1u) {
    p = eval(expr.substr(0, pos));
    typestring = Eval_Type;
    q = eval(expr.substr(pos + operation.length()));
    typestring += Eval_Type;
    if (typestring == "NN") {
      pb = new bool;
      switch (Field) {
        case RAT: pnq = (Q*)p;
                  qnq = (Q*)q;
                  if (operation == "<=")
                    *pb = *pnq <= *qnq;
                  else if (operation == "<")
                    *pb = *pnq < *qnq;
                  else if (operation == ">=")
                    *pb = *pnq >= *qnq;
                  else if (operation == ">")
                    *pb = *pnq > *qnq;
                  delete pnq, qnq;
                  Eval_Type = 'B';
                  break;
        case MOD: pnf = (Fp*)p;
                  qnf = (Fp*)q;
                  if (operation == "<=")
                    *pb = *pnf <= *qnf;
                  else if (operation == "<")
                    *pb = *pnf < *qnf;
                  else if (operation == ">=")
                    *pb = *pnf >= *qnf;
                  else if (operation == ">")
                    *pb = *pnf > *qnf;
                  delete pnf, qnf;
                  Eval_Type = 'B';
                  break;
      }
      return (int*)pb;
    }
    else if (typestring == "NP") {
      pb = new bool;
      switch (Field) {
        case RAT: pnq = (     Q *)p;
                  qpq = (poly<Q>*)q;
                  PPQ = poly<Q>(*pnq);
                  if (operation == "<=")
                    *pb = PPQ <= *qpq;
                  else if (operation == "<")
                    *pb = PPQ < *qpq;
                  else if (operation == ">=")
                    *pb = PPQ >= *qpq;
                  else if (operation == ">")
                    *pb = PPQ > *qpq;
                  delete pnq, qpq;
                  Eval_Type = 'B';
                  break;
        case MOD: pnf = (     Fp *)p;
                  qpf = (poly<Fp>*)q;
                  PPF = poly<Fp>(*pnf);
                  if (operation == "<=")
                    *pb = PPF <= *qpf;
                  else if (operation == "<")
                    *pb = PPF < *qpf;
                  else if (operation == ">=")
                    *pb = PPF >= *qpf;
                  else if (operation == ">")
                    *pb = PPF > *qpf;
                  delete pnf, qpf;
                  Eval_Type = 'B';
                  break;
      }
      return (int*)pb;
    }
    else if (typestring == "PN") {
      pb = new bool;
      switch (Field) {
        case RAT: ppq = (poly<Q>*)p;
                  qnq = (     Q *)q;
                  QPQ = poly<Q>(*qnq);
                  if (operation == "<=")
                    *pb = *ppq <= QPQ;
                  else if (operation == "<")
                    *pb = *ppq < QPQ;
                  else if (operation == ">=")
                    *pb = *ppq >= QPQ;
                  else if (operation == ">")
                    *pb = *ppq > QPQ;
                  delete ppq, qnq;
                  Eval_Type = 'B';
                  break;
        case MOD: ppf = (poly<Fp>*)p;
                  qnf = (     Fp *)q;
                  QPF = poly<Fp>(*qnf);
                  if (operation == "<=")
                    *pb = *ppf <= QPF;
                  else if (operation == "<")
                    *pb = *ppf < QPF;
                  else if (operation == ">=")
                    *pb = *ppf >= QPF;
                  else if (operation == ">")
                    *pb = *ppf > QPF;
                  delete ppf, qnf;
                  Eval_Type = 'B';
                  break;
      }
      return (int*)pb;
    }
    else if (typestring == "PP") {
      pb = new bool;
      switch (Field) {
        case RAT: ppq = (poly<Q>*)p;
                  qpq = (poly<Q>*)q;
                  if (operation == "<=")
                    *pb = *ppq <= *qpq;
                  else if (operation == "<")
                    *pb = *ppq < *qpq;
                  else if (operation == ">=")
                    *pb = *ppq >= *qpq;
                  else if (operation == ">")
                    *pb = *ppq > *qpq;
                  delete ppq, qpq;
                  Eval_Type = 'B';
                  break;
        case MOD: ppf = (poly<Fp>*)p;
                  qpf = (poly<Fp>*)q;
                  if (operation == "<=")
                    *pb = *ppf <= *qpf;
                  else if (operation == "<")
                    *pb = *ppf < *qpf;
                  else if (operation == ">=")
                    *pb = *ppf >= *qpf;
                  else if (operation == ">")
                    *pb = *ppf > *qpf;
                  delete ppf, qpf;
                  Eval_Type = 'B';
                  break;
      }
      return (int*)pb;
    }
    //else if (typestring == "NI" && operation[0] == '<') {
    //}
    //else if (typestring == "IN" && operation[0] == '>') {
    //}
    //else if (typestring == "PI" && operation[0] == '<') {
    //}
    //else if (typestring == "IP" && operation[0] == '>') {
    //}
    //else if (typestring == "II") {
    //}
    else if (typestring == "NG" && operation[0] == '<') {
      pb = new bool;
      switch (Field) {
        case RAT: pnq = (          Q   *)p;
                  qgq = (list<poly<Q> >*)q;
                  *pb = poly<Q>(*pnq) < *qgq;
                  delete pnq, qgq;
                  Eval_Type = 'B';
                  break;
        case MOD: pnf = (          Fp   *)p;
                  qgf = (list<poly<Fp> >*)q;
                  *pb = poly<Fp>(*pnf) < *qgf;
                  delete pnf, qgf;
                  Eval_Type = 'B';
                  break;
      }
      return (int*)pb;
    }
    else if (typestring == "PG" && operation[0] == '<') {
      pb = new bool;
      switch (Field) {
        case RAT: ppq = (     poly<Q>  *)p;
                  qgq = (list<poly<Q> >*)q;
                  *pb = *ppq < *qgq;
                  delete ppq, qgq;
                  Eval_Type = 'B';
                  break;
        case MOD: ppf = (     poly<Fp>  *)p;
                  qgf = (list<poly<Fp> >*)q;
                  *pb = *ppf < *qgf;
                  delete ppf, qgf;
                  Eval_Type = 'B';
                  break;
      }
      return (int*)pb;
    }
    else {
      release(p, typestring[0]);
      release(q, typestring[1]);
      Eval_Type = '-';
      return NULL;
    }
  }

  // search for "+", "-"
  pos = -1u;
  if (symbols.length() > 1) {
    pos = evalfindstring(symbols.substr(1), "+");
    if (pos == -1u) {
      pos = evalfindstring(symbols.substr(1), "-");
      if (pos != -1u) {
        operation = "-";
        pos++;
      }
    }
    else {
      operation = "+";
      pos++;
    }
  }
  if (pos != -1u) {
    p = eval(expr.substr(0, pos));
    typestring = Eval_Type;
    if (operation == "-") {
      operation = "+";
      pos--;
    }
    q = eval(expr.substr(pos + 1));
    typestring += Eval_Type;
    if (typestring == "NN") {
      switch (Field) {
        case RAT: pnq = (Q*)p;
                  qnq = (Q*)q;
                  *pnq += *qnq;
                  delete qnq;
                  return (int*)pnq;
        case MOD: pnf = (Fp*)p;
                  qnf = (Fp*)q;
                  *pnf += *qnf;
                  delete qnf;
                  return (int*)pnf;
      }
    }
    else if (typestring == "NP") {
      switch (Field) {
        case RAT: pnq = (     Q *)p;
                  qpq = (poly<Q>*)q;
                  *qpq += *pnq;
                  delete pnq;
                  return (int*)qpq;
        case MOD: pnf = (     Fp *)p;
                  qpf = (poly<Fp>*)q;
                  *qpf += *pnf;
                  delete pnf;
                  return (int*)qpf;
      }
    }
    else if (typestring == "PN") {
      switch (Field) {
        case RAT: ppq = (poly<Q>*)p;
                  qnq = (     Q *)q;
                  *ppq += *qnq;
                  delete qnq;
                  Eval_Type = 'P';
                  return (int*)ppq;
        case MOD: ppf = (poly<Fp>*)p;
                  qnf = (     Fp *)q;
                  *ppf += *qnf;
                  delete qnf;
                  return (int*)ppf;
      }
    }
    else if (typestring == "PP") {
      switch (Field) {
        case RAT: ppq = (poly<Q>*)p;
                  qpq = (poly<Q>*)q;
                  *ppq += *qpq;
                  delete qpq;
                  return (int*)ppq;
        case MOD: ppf = (poly<Fp>*)p;
                  qpf = (poly<Fp>*)q;
                  *ppf += *qpf;
                  delete qpf;
                  return (int*)ppf;
      }
    }
    else if (typestring == "II") {
      switch (Field) {
        case RAT: piq = (ideal<Q>*)p;
                  qiq = (ideal<Q>*)q;
                  *piq += *qiq;
                  delete qiq;
                  return (int*)piq;
        case MOD: pif = (ideal<Fp>*)p;
                  qif = (ideal<Fp>*)q;
                  *pif += *qif;
                  delete qif;
                  return (int*)pif;
      }
    }
    else if (typestring == "GG") {
      switch (Field) {
        case RAT: pgq = (list<poly<Q> >*)p;
                  qgq = (list<poly<Q> >*)q;
                  *pgq += *qgq;
                  delete qgq;
                  return (int*)pgq;
        case MOD: pgf = (list<poly<Fp> >*)p;
                  qgf = (list<poly<Fp> >*)q;
                  *pgf += *qgf;
                  delete qgf;
                  return (int*)pgf;
      }
    }
    else {
      release(p, typestring[0]);
      release(q, typestring[1]);
      Eval_Type = '-';
      return NULL;
    }
  }

  // search for "*", "/"
  pos = evalfindstring(symbols, "*");
  if (pos == -1u) {
    pos = evalfindstring(symbols, "/");
    if (pos != -1u)
      operation = "/";
  }
  else
    operation = "*";
  if (pos != -1u) {
    p = eval(expr.substr(0, pos));
    typestring = Eval_Type;
    q = eval(expr.substr(pos + operation.length()));
    typestring += Eval_Type;
    if (typestring == "NN") {
      switch (Field) {
        case RAT: pnq = (Q*)p;
                  qnq = (Q*)q;
                  if (operation == "*")
                    *pnq *= *qnq;
                  else {
                    if (*qnq == 0) {
                      delete pnq, qnq;
                      Eval_Type = '-';
                      return NULL;
                    }
                    *pnq /= *qnq;
                  }
                  delete qnq;
                  return (int*)pnq;
        case MOD: pnf = (Fp*)p;
                  qnf = (Fp*)q;
                  if (operation == "*")
                    *pnf *= *qnf;
                  else {
                    if (qnf == 0) {
                      delete pnf, qnf;
                      Eval_Type = '-';
                      return NULL;
                    }
                    *pnf /= *qnf;
                  }
                  delete qnf;
                  return (int*)pnf;
      }
    }
    else if (typestring == "NP") {
      switch (Field) {
        case RAT: pnq  = (     Q *)p;
                  qpq  = (poly<Q>*)q;
                  if (operation == "*") {
                    *qpq *= *pnq;
                    delete pnq;
                    return (int*)qpq;
                  }
                  delete pnq, qpq;
                  break;
        case MOD: pnf  = (     Fp *)p;
                  qpf  = (poly<Fp>*)q;
                  if (operation == "*") {
                    *qpf *= *pnf;
                    delete pnf;
                    return (int*)qpf;
                  }
                  delete pnf, qpf;
                  break;
      }
    }
    else if (typestring == "PN") {
      switch (Field) {
        case RAT: ppq = (poly<Q>*)p;
                  qnq = (     Q *)q;
                  if (operation == "*")
                    *ppq *= *qnq;
                  else {
                    if (*qnq == 0) {
                      delete ppq, qnq;
                      Eval_Type = '-';
                      return NULL;
                    }
                    *ppq /= *qnq;
                  }
                  delete qnq;
                  Eval_Type = 'P';
                  return (int*)ppq;
        case MOD: ppf = (poly<Fp>*)p;
                  qnf = (     Fp *)q;
                  if (operation == "*")
                    *ppf *= *qnf;
                  else {
                    if (*qnf == 0) {
                      delete ppf, qnf;
                      Eval_Type = '-';
                      return NULL;
                    }
                    *ppf /= *qnf;
                  }
                  delete qnf;
                  Eval_Type = 'P';
                  return (int*)ppf;
      }
    }
    else if (typestring == "PP") {
      switch (Field) {
        case RAT: ppq = (poly<Q>*)p;
                  qpq = (poly<Q>*)q;
                  if (operation == "*")
                    *ppq *= *qpq;
                  else {
                    if (*qpq == poly<Q>(0)) {
                      delete ppq, qpq;
                      Eval_Type = '-';
                      return NULL;
                    }
                    *ppq /= *qpq;
                  }
                  delete qpq;
                  return (int*)ppq;
        case MOD: ppf = (poly<Fp>*)p;
                  qpf = (poly<Fp>*)q;
                  if (operation == "*")
                    *ppf *= *qpf;
                  else {
                    if (*qpf == poly<Fp>(0)) {
                      delete ppf, qpf;
                      Eval_Type = '-';
                      return NULL;
                    }
                    *ppf /= *qpf;
                  }
                  delete qpf;
                  return (int*)ppf;
      }
    }
    else if (typestring == "II") {
      switch (Field) {
        case RAT: piq = (ideal<Q>*)p;
                  qiq = (ideal<Q>*)q;
                  if (operation == "*") {
                    *piq *= *qiq;
                    delete qiq;
                    return (int*)piq;
                  }
                  delete piq, qiq;
                  Eval_Type = '-';
                  return NULL;
        case MOD: pif = (ideal<Fp>*)p;
                  qif = (ideal<Fp>*)q;
                  if (operation == "*") {
                    *pif *= *qif;
                    delete qif;
                    return (int*)pif;
                  }
                  delete pif, qif;
                  Eval_Type = '-';
                  return NULL;
      }
    }
    else {
      release(p, typestring[0]);
      release(q, typestring[1]);
      Eval_Type = '-';
      return NULL;
    }
  }

  // search for "^"
  pos = evalfindstring(symbols, "^");
  if (pos != -1u) {
    Exp_Depth++;
    q = eval(expr.substr(pos + 1));
    Exp_Depth--;
    if (Eval_Type == 'N') {
      Z exponent;
      qnq = (Q*)q;
      if ((*qnq).den == 1) {
        exponent = (*qnq).num;
        delete qnq;
        p = eval(expr.substr(0, pos));
        switch (Eval_Type) {
          case 'N': switch (Field) {
                      case RAT: pnq  = (Q*)p;
                                *pnq ^= exponent;
                                return (int*)pnq;
                      case MOD: pnf  = (Fp*)p;
                                *pnf ^= exponent;
                                return (int*)pnf;
                    }
          case 'P': switch (Field) {
                      case RAT: ppq  = (poly<Q>*)p;
                                *ppq ^= exponent;
                                return (int*)ppq;
                      case MOD: ppf  = (poly<Fp>*)p;
                                *ppf ^= exponent;
                                return (int*)ppf;
                    }
        }
      }
      else {
        release(p, Eval_Type);
        delete qnq;
        Eval_Type = '-';
        return NULL;
      }
    }
    else
      release(q, Eval_Type);
  }

  Eval_Type = '-';
  return NULL;
}

// find operators not enclosed in brackets

unsigned int evalfindstring(const string &s, const string &op) {
  unsigned int brackets = 0, i, j, ls = s.length(), lop = op.length();
  for (i = 0; i < ls; i++)
    switch (s[i]) {
      case '(': brackets++;
                break;
      case ')': if (brackets == 0)
                  return -1u;
                brackets--;
                break;
      default : if (brackets == 0 && i + lop < ls) {
                  for (j = 0; j < lop && s[i + j] == op[j]; j++);
                  if (j == lop)
                    return i;
                }
    }
  return -1u;
}

// check input

void evaluate(string orgline) {
  string command, command2, parseline;
  unsigned long c, i, n = counttokens(orgline, ';');
  bool success = true;

  for (c = 1; c <= n; c++) {
    parseline = gettoken(orgline, ';', c);
    if (parseline == "")
      continue;

    command = getdefinition(parseline);
    if (command != "") {
      assignvariable(command, parseline);
      continue;
    }

    command = lower(getfirstword(parseline));

    if (command == "bye" || command == "exit" || command == "quit")
      Parser_Quit = true;

    else if (command == "define")
      assignvariable(getfirstword(parseline), parseline);

    else if (command == "help")
      help(parseline);

    else if (command == "print" || command == "?")
      display(parseline);

    else if (command == "run")
      runprogram(parseline);

    else if (command == "set")
      setvar(parseline);

    else if (command == "undefine")
      deletevariable(getfirstword(parseline));

    else if (command == "use") {
      i = counttokens(parseline, ',');
      if (i == 0) {
        cout << "Please give a polynomial ring to set!" << endl << endl;
        help("use");
        success = false;
      }
      else {
        if (i > 1) {
          command2  = lower(trim(gettoken(parseline, ',', i)));
          if (command2 == "lex" || command2 == "deglex" || command2 == "degrevlex" || command2.substr(0, 4) == "elim")
            parseline = parseline.substr(0, parseline.length() - command2.length() - 1);
          else
            command2 = "";
        }
        command  = "ring "  ;
        command += parseline;
        if (setvar(command) && command2 != "") {
          command  = "termord ";
          command += command2  ;
          setvar(command);
        }
      }
    }

    else if (command == "varlist")
      varlist();

    else {
      cout << "Unknown command \"" << upper(command) << "\"!" << endl;
      success = false;
    }

    if (!success)
      break;
  }
}


// retrieve text in brackets ( [ { <

string getbrackets(string &s, const string &b) {
  s = trim(s);

  string bracketstring;
  char   left, right, c;
  unsigned short depth = 1, i = 1, l = s.length();

  if (b == "") {
    left = s[0];
    bracketstring = left;
  }
  else
    left = b[0];
  switch (left) {
    case '(': right = ')';
              break;
    case '[': right = ']';
              break;
    case '{': right = '}';
              break;
    case '<': right = '>';
              break;
    default : return "";
  }
  if (s[0] != left)
    return "";

  for (; i < l; i++) {
    c = s[i];
    if (c == left) {
      ++depth;
      bracketstring += c;
    }
    else
      if (c == right) {
        if (--depth == 0)
          break;
        else
          bracketstring += c;
      }
      else
        bracketstring += c;
  }

  if (b == "")
    bracketstring += right;
  return bracketstring;
}


// retrieve parts of a definition including :=

string getdefinition(string &s) {
  s = trim(s);

  unsigned int defpos = s.find(":=");
  string def;

  if (defpos != -1u) {
    def = trim(s.substr(0, defpos ));
    s   = trim(s.substr(defpos + 2));
    if (def == "")
      return " ";
  }
  return def;
}


// retrieve first word (alphanum, but nut the first letter)

string getfirstword(string &s) {
  s = trim(s);

  string head = "";
  unsigned short i = 0, l = s.length();

  if (isalpha(s[0])) {
    for (; i < l; i++)
      if (isalnum(s[i]) || s[i] == '_' || s[i] == '?')
        head += s[i];
      else
        break;
    s = ltrim(s.substr(i));
  }
  else
    if (s[0] == '?') {
      s = ltrim(s.substr(1));
      return "?";
    }
  return head;
}


// retrieve indeterminate

string getindet(string &s) {
  s = trim(s);

  unsigned long pos;
  string bracket, old = s, sub;

  sub      = s[0];
  s        = s.substr(1);
  bracket  = getbrackets(s, "[");
  if (bracket == "")
    pos = IndetList.searchex(sub[0]);
  else {
    pos = IndetList.search(sub[0], val(bracket));
    sub += '['    ;
    sub += bracket;
    sub += ']'    ;
  }
  if (pos) {
    if (sub == old)
      s = "";
    else
      s = old.substr(sub.length());
    return sub;
  }
  s = old;
  return "";
}


// retrieve first number

string getnumber(string &s) {
  s = trim(s);

  string head = "";
  unsigned short i = 0, l = s.length();

  for (; i < l; i++)
    if (isdigit(s[i]))
      head += s[i];
    else
      break;
  s = ltrim(s.substr(i));
  return head;
}


// retrieve first operator + - * / % = : < >

string getoperator(string &s) {
  s = trim(s);

  if (s == "")
    return "";

  string op = "";
  char c = s[0];

  if (c == '+' || c == '-' || c == '*' || c == '/' || c == '%' || c == '=' || c == ':' || c == '<' || c == '>') {
    s = ltrim(s.substr(1));
    op = c;
  }
  return op;
}


// retriev first punctation . , : ; ? ! -

string getpunctuation(string &s) {
  s = trim(s);

  if (s == "")
    return "";

  string punct = "";
  char c = s[0];

  if (c == '.' || c == ',' || c == ':' || c == ';' || c == '?' || c == '!' || c == '-') {
    s = ltrim(s.substr(1));
    punct = c;
  }
  return punct;
}


// help text

void help(string topic) {
  topic = lower(getfirstword(topic));
  cout << endl;

  if (topic == "")
    cout << "List of all available commands:"                                                << endl
         <<                                                                                     endl
         << "BYE       EXIT      PRINT     RUN       UNDEFINE  VARLIST"                      << endl
         << "DEFINE    HELP      QUIT      SET       USE       ?"                            << endl
         <<                                                                                     endl
         <<                                                                                     endl
         << "List of all available functions:"                                               << endl
         <<                                                                                     endl
         << "GB        IDEAL     LEN       POLY"                                             << endl
         <<                                                                                     endl
         <<                                                                                     endl
         << "Both commands and functions are not case sensitive."                            << endl
         << "Type \"HELP <command>\" or \"HELP <function>\" for detailled help information." << endl;

  else if (topic == "bye" || topic == "exit" || topic == "quit")
    cout << "syntax: BYE"                                              << endl
         << "        EXIT"                                             << endl
         << "        QUIT"                                             << endl
         <<                                                               endl
         << "description: All of these commands quit GBC immediately." << endl;

  else if (topic == "define")
    cout << "syntax: DEFINE <variable> <expression>"             << endl
         << "    or: <variable> := <expression>"                 << endl
         <<                                                         endl
         << "description: Defines or redefines a user variable." << endl
         <<                                                         endl
         << "examples: DEFINE I Ideal(x)"                        << endl
         << "          I := Ideal(x, y)"                         << endl;

  else if (topic == "gb")
    cout << "syntax: GB(I [, alg])"                                                       << endl
         <<                                                                                  endl
         << "parameters: I   - ideal"                                                     << endl
         << "            alg - algorithm shortcut (cf. \"HELP GBALG\")"                   << endl
         <<                                                                                  endl
         << "return type: list"                                                           << endl
         <<                                                                                  endl
         << "description: Calculates a Grbner basis from an ideal I. If no shortcut is"  << endl
         << "             given, the algorithm set with \"SET GBALG\" will be used."      << endl
         <<                                                                                  endl
         << "examples: PRINT GB(IDEAL(x, 2y))"                                            << endl
         << "          I := Ideal(x, 2y^2 + 3y, z]; PRINT GB(I, Buchberger)"              << endl;

  else if (topic == "gbalg")
    cout << "List of available Grbner basis algorithm shortcuts:"                         << endl
         <<                                                                                   endl
         << "Buchberger algorithm:                         B, Buch, Buchberger"            << endl
         << "for homogeneous ideals:                       BH, BuchHomo, BuchbergerHomo"   << endl
         << "for homogeneous ideals using Gauss:           BG, BuchGauss, BuchbergerGauss" << endl
         << "for homogeneous ideals using multi reduction: BM, BuchMulti, BuchbergerMulti" << endl
         << "F4 for homogeneous ideals:                    F4, Fau, Faugere"               << endl
         <<                                                                                   endl
         << "These are not case sensitive and are used with \"GB\" and \"SET GBALG\"."     << endl;

  else if (topic == "gbopt")
    cout << "List of available optimization shortcuts:"                                        << endl
         <<                                                                                       endl
         << "If you do not specify \"off\" (for no optimization), GBC will scan the parameter" << endl
         << "string for these letters:"                                                        << endl
         <<                                                                                       endl
         << "N: Normal Selection Strategy"                                                     << endl
         << "G: Gebauer Mller"                                                                << endl
         << "P: Pair Minimalisation"                                                              << endl
         <<                                                                                       endl
         << "These are not case sensitive and are used with \"SET GBOPT\"."                    << endl
         << "There are some restrictions:"                                                     << endl
         << "- K works only for homogeneous algorithms."                                       << endl
         << "- G and K cannot be used together."                                               << endl
         << "If this is not satisfied, GBC will ignore some letters in order to get proper"    << endl
         << "settings."                                                                        << endl
         <<                                                                                       endl
         << "GBC will select a proper optimazation when setting a Grbner basis algorithm,"    << endl
         << "i.e. you only have to use \"SET GBOPT\" if you prefer a different one."           << endl;

  else if (topic == "help")
    cout << "syntax: HELP [command]"                                                              << endl
         <<                                                                                          endl
         << "description: Shows some helpful information on a command or -"                       << endl
         << "             if you did not give a command - a list of all available commands."      << endl
         <<                                                                                          endl
         << "example: HELP HELP"                                                                  << endl
         << "         That is what you typed to get this. - Really!"                              << endl;

  else if (topic == "ideal")
    cout << "syntax: IDEAL([p1 [, p2 [,...]]])"                                              << endl
         << "        IDEAL(L)"                                                               << endl
         <<                                                                                     endl
         << "parameters: p1, p2,... - polynomials"                                           << endl
         << "            L          - list of polynomials"                                   << endl
         <<                                                                                     endl
         << "return type: ideal"                                                             << endl
         <<                                                                                     endl
         << "description: Defines an ideal with generator system p1, p2,..."                 << endl
         << "             In the case you give a list, it will be used as generator system." << endl
         <<                                                                                     endl
         << "examples: PRINT IDEAL()"                                                        << endl
         << "          PRINT IDEAL(x, y^2)"                                                  << endl
         << "          L := [y, 2x, z]; PRINT IDEAL(L)"                                      << endl;

  else if (topic == "len")
    cout << "syntax: LEN(L)"                             << endl
         <<                                                 endl
         << "parameter: L - list"                        << endl
         <<                                                 endl
         << "return type: number"                        << endl
         <<                                                 endl
         << "description: Returns the length of a list." << endl
         <<                                                 endl
         << "examples: PRINT LEN([])"                    << endl
         << "          L := [x, y, z]; PRINT LEN(L)"     << endl;

  else if (topic == "poly")
    cout << "syntax: POLY(n)"                                                     << endl
         <<                                                                          endl
         << "parameter: n - field element"                                        << endl
         <<                                                                          endl
         << "return type: polynomial"                                             << endl
         <<                                                                          endl
         << "description: Casts a field element n to a polynomial."               << endl
         <<                                                                          endl
         << "example: PRINT POLY(2)"                                              << endl
         << "         Displays the polynomial \"2\""                              << endl
         << "         (which is optically the same as if you typed \"PRINT 2\")." << endl
         <<                                                                          endl
         << "note: Normally GBC does casts automatically."                        << endl;

  else if (topic == "print" || topic == "?")
    cout << "syntax: PRINT <expression>"                          << endl
         << "    or: ?     <expression>"                          << endl
         <<                                                          endl
         << "description: Displays the content of an expression." << endl
         <<                                                          endl
         << "examples: PRINT a"                                   << endl
         << "          ? 1 + 3"                                   << endl;

  else if (topic == "run")
    cout << "syntax: RUN <program> [WITH ...#...]"                                               << endl
         <<                                                                                         endl
         << "description: Runs a GBC program stored in a file with the extension \".gbc\","      << endl
         << "             writing the output into a file with same name and extension \".out\"."
         << "             The WITH clause allows you to modify the program lines. RUN will"      << endl
         << "             create lines looking like what is written after WITH, replacing all"
         << "             occurences of # with the original program line."                       << endl
         << "             By this way, you do not have to repeat commands you want to execute"
         << "             for every line. For example, you could put a list of ideals into a"    << endl
         << "             program file and calculate their Grbner bases without modifying"      << endl
         << "             the file (cf. second example)."                                        << endl
         << "             You can add comments to your programs by adding a double slash"        << endl
         << "             (\"//\") at the beginning of each comment line."                       << endl
         <<                                                                                         endl
         << "examples: RUN testprog"                                                             << endl
         << "          RUN testprog WITH PRINT GB(#)"                                            << endl
         << "          RUN testprog WITH PRINT 2 * (# + #)"                                      << endl;

  else if (topic == "set")
    cout << "syntax: SET <envvar> [value]"                                                     << endl
         <<                                                                                       endl
         << "description: Sets a GBC environment variable to a given value."                   << endl
         << "             If no value is given the actual value will be displayes."            << endl
         <<                                                                                       endl
         << "available GBC environment variables are:"                                         << endl
         << "Ŀ"
         << " variable  allowed settings                                                  "
         << "Ĵ"
         << " GBALG     a Grbner basis algorithm shortcut - type \"HELP GBALG\" for a list "
         << " GBOPT     an optimization shortcut - type \"HELP GBOPT\" for a list           "
         << " RING      a field, followed by a list of indeterminates, eg. Q[X], Z/2[x,y] "
         << " STATS     ON or OFF - use this to switch the display of program statistics  "
         << " TERMORD   a term ordering: Deg, DegLex, DegRevLex, Elim(n)                  "
         << ""
         <<                                                                                       endl
         << "examples: SET RING Q[x]"                                                          << endl
         << "          This sets the polynomial ring to use to Q[x]."                          << endl
         <<                                                                                       endl
         << "          SET RING"                                                               << endl
         << "          This displays the polynomial ring which is currently in use."           << endl;

  else if (topic == "undefine")
    cout << "syntax: UNDEFINE <variable>"                       << endl
         <<                                                        endl
         << "description: Deletes a user variable from memory." << endl
         <<                                                        endl
         << "example: UNDEFINE I"                               << endl;

  else if (topic == "use")
    cout << "syntax: USE <ring> [, <termord>]"                                               << endl
         <<                                                                                     endl
         << "description: USE sets the active ring and optionally the active term ordering." << endl
         << "             If no term ordering is given, it will not be changed."             << endl
         <<                                                                                     endl
         << "examples: USE Q[x, y]"                                                          << endl
         << "          USE Z/3[x[1..3]], DegRevLex"                                          << endl
         <<                                                                                     endl
         << "note: This is the same like using SET RING and SET TERMORD"                     << endl;

  else if (topic == "varlist")
    cout << "syntax: VARLIST"                                               << endl
         <<                                                                    endl
         << "description: Displays a list of all user variables in memory." << endl;

  else cout << "Unknown help topic \"" << upper(topic) << "\"!" << endl;
}


// read an input line

string readline() {
  string inputline;
  cout << ">> ";
  getline(cin, inputline);
  return inputline;
}


// releasing a temporary variable (used for evaluation)

void release(const int* &p, const char &type) {
  bool             *pb ;
  Q                *pnq;
  Fp               *pnf;
  poly <     Q   > *ppq;
  poly <     Fp  > *ppf;
  ideal<     Q   > *piq;
  ideal<     Fp  > *pif;
  list <poly<Q > > *pgq;
  list <poly<Fp> > *pgf;

  switch (type) {
    case 'B': pb = (bool*)p;
              delete pb;
              return;
    case 'N': switch (Field) {
                case RAT: pnq = (Q*)p;
                          delete pnq;
                          return;
                case MOD: pnf = (Fp*)p;
                          delete pnf;
                          return;
              }
    case 'P': switch (Field) {
                case RAT: ppq = (poly<Q>*)p;
                          delete ppq;
                          return;
                case MOD: ppf = (poly<Fp>*)p;
                          delete ppf;
                          return;
              }
    case 'I': switch (Field) {
                case RAT: piq = (ideal<Q>*)p;
                          delete piq;
                          return;
                case MOD: pif = (ideal<Fp>*)p;
                          delete pif;
                          return;
              }
    case 'G': switch (Field) {
                case RAT: pgq = (list<poly<Q> >*)p;
                          delete pgq;
                          return;
                case MOD: pgf = (list<poly<Fp> >*)p;
                          delete pgf;
                          return;
              }
  }
}


// display a GBC setting

void runprogram(string params) {
  if (params =="") {
    cout << "Please give the filename of the GBC program to open." << endl;
    return;
  }

  unsigned int pos = evalfindstring(upper(params), " WITH");
  bool modi = pos != -1lu;
  string infile, outfile, inputline;
  ifstream fin ;
  ofstream fout;
  istream  sin  = cin ;
  ostream  sout = cout;

  if (modi) {
    infile = left(params, pos);
    params = trim(params.substr(pos + 5));
  }
  else {
    infile = params;
    params = "";
  }
  outfile = infile + ".out";
  infile += ".gbc";

  fin.open(infile.c_str(), ios::binary);
  if (!fin) {
    cout << "File \"" << infile << "\" cannot be opened." << endl;
    return;
  }
  fout.open(outfile.c_str(), ios::binary);
  if (!fout) {
    cout << "Output file \"" << outfile << "\" cannot be opened." << endl;
    fin.close();
    return;
  }

  cin  = fin ;
  cout = fout;

  while (!cin.eof() && !Parser_Quit) {
    getline(cin, inputline);
    if (!cin.eof()) {
      inputline = left(inputline, inputline.length() - 1);
      if (inputline != "" && left(inputline, 2) != "//") {
        if (modi)
          inputline = repsubstring(params, "#", inputline);
        cout << ">> " << inputline << endl;
        evaluate(inputline);
        cout << endl;
      }
    }
  }

  fin.close();
  fout.close();
  cin  = sin ;
  cout = sout;

  fin.open(outfile.c_str(), ios::binary);
  if (fin) {
    while (!fin.eof()) {
      getline(fin, inputline);
      if (!fin.eof())
        cout << endl << inputline;
    }
    fin.close();
    cout << endl << "End of program " << infile << endl;
  }
}

// display a GBC setting

void setting(string envvar) {
  envvar = lower(getfirstword(envvar));

  if (envvar == "") {
    help("");
    return;
  }

  if (envvar == "gbalg") {
    string alg;
    switch (GBType) {
      case Buchberger     : alg = "Buchberger";
                            break;
      case BuchbergerHomo : alg = "BuchbergerHomo";
                            break;
      case BuchbergerGauss: alg = "BuchbergerGauss";
                            break;
      case BuchbergerMulti: alg = "BuchbergerMulti";
                            break;
      case F4             : alg = "F4";
                            break;
    }
    cout << "The default Grbner basis algorithm is \"" << alg << "\"." << endl;
    return;
  }

  if (envvar == "gbopt") {
    cout << "The active optimization is \"" << GetOpt() << "\" (" << GetOptString() << ")." << endl;
    return;
  }

  if (envvar == "ring") {
    string field;
    switch (Field) {
      case RAT: field += "Q";
                break;
      case MOD: field += "Z/"    ;
                field += Fp_prime;
                break;
    }
    field += '['             ;
    field += indets.GetList();
    field += ']'             ;
    cout << "The active ring is " << field << '.' << endl;
    return;
  }

  if (envvar == "stats") {
    cout << "STATS is " << (Show_Stats ? "ON" : "OFF") << '.' << endl;
    return;
  }

  if (envvar == "termord") {
    string to;
    switch (TermOrd) {
      case Lex      : to = "Lex";
                      break;
      case DegLex   : to = "DegLex";
                      break;
      case DegRevLex: to = "DegRevLex";
                      break;
      case Elim     : to  = "Elim("    ;
                      to += TermOrdElim;
                      to += ")"        ;
    }
    cout << "The active term ordering is " << to << '.' << endl;
    return;
  }

  cout << "Unkown setting \"" << upper(envvar) << "\"!" << endl;
}


// set a GBC setting

bool setvar(string value) {
  string oldval     = value                     ,
         envvar     = lower(getfirstword(value)),
         givenvalue = value                     ;

  if (envvar == "") {
    cout << "Please give an GBC environment variable to set!" << endl << endl;
    help("set");
    return false;
  }

  if (givenvalue == "") {
    setting(oldval);
    return true;
  }

  if (envvar == "gbalg") {
    string alg = lower(getfirstword(value));
    if (alg == "b" || alg == "buch" || alg == "buchberger") {
      alg = "Buchberger";
      SetGBAlg("Buchberger");
      SetOpt("ng");
    }
    else if (alg == "bh" || alg == "buchhomo" || alg == "buchbergerhomo") {
      alg = "BuchbergerHomo";
      SetGBAlg("BuchbergerHomo");
      SetOpt("np");
    }
    else if (alg == "bg" || alg == "buchgauss" || alg == "buchbergergauss") {
      alg = "BuchbergerGauss";
      SetGBAlg("BuchbergerGauss");
      SetOpt("np");
    }
    else if (alg == "bm" || alg == "buchmulti" || alg == "buchbergermulti") {
      alg = "BuchbergerMulti";
      SetGBAlg("BuchbergerMulti");
      SetOpt("p");
    }
    else if (alg == "f4" || alg == "fau" || alg == "faugere") {
      alg = "F4";
      SetGBAlg("F4");
      SetOpt("p");
    }
    else {
      cout << "\"" << givenvalue << "\" is not a valid Grbner basis algorithm shortcut" << endl;
      help("GBALG");
      return false;
    }
    cout << "(OK) The default Grbner basis algorithm was set to \"" << alg << "\"." << endl;
    return true;
  }

  if (envvar == "gbopt") {
    SetOpt(lower(getfirstword(value)));
    cout << "(OK) The optimization was set to \"" << GetOpt() << "\" (" << GetOptString() << ")." << endl;
    return true;
  }

  if (envvar == "ring") {
    string bracketstring                      ,
           field         = getfirstword(value),
           indetstring                        ,
           indetrange                         ,
           indetname                          ,
           msg                                ,
           nr                                 ,
           op                                 ,
           punct1                             ,
           punct2                             ;
    unsigned int i, indetcount;
    unsigned long first;
    Z p;
    FIELD newfield = none;
    bool indetfound = false;
    if (field == "") {
      cout << "Please give a polynomial ring to set!" << endl << endl;
      help("set");
      return false;
    }
    if (field == "Q") {
      newfield = RAT;
      msg      = "Q";
    }
    if (field == "Z") {
      op   = getoperator(value);
      nr   = getnumber(value);
      p    = val(nr);
      msg  = "Z/";
      msg += nr  ;
      if (op == "/" && IsPrime(p))
        newfield = MOD;
      else {
        cout << msg << " is not a field!" << endl;
        return false;
      }
    }
    bracketstring = getbrackets(value, "[");
    if (newfield == none || bracketstring == "") {
      cout << "\"" << givenvalue << "\" is not a valid polynomial ring!" << endl;
      return false;
    }
    indetcount = counttokens(bracketstring, ',');
    for (i = 1; i <= indetcount; i++) {
      indetstring = gettoken(bracketstring, ',', i);
      indetname   = getfirstword(indetstring);
      if (indetname.length() == 1) {
        if (!indetfound) {
          indetfound = true;
          indets.removeall();
        }
        if (!IndetList.search(indetname[0])) {
          indetrange = getbrackets(indetstring, "[");
          nr         = getnumber(indetrange);
          if (nr == "")
            indets.add(indetname[0]);
          else {
            first  = val(nr)                   ;
            punct1 = getpunctuation(indetrange);
            punct2 = getpunctuation(indetrange);
            nr     = getnumber(indetrange)     ;
            if (punct1 != "." || punct2 != "." || nr == "")
              indets.add(indetname[0], first);
            else
              indets.add(indetname[0], first, val(nr));
          }
        }
        else
          cout << "- Removed double indeterminate: " << indetname << indetstring << endl;
      }
    }
    if (indetfound) {
      Field = newfield;
      if (field == "Z")
        SetFpPrime(p);
      msg += '['             ;
      msg += indets.GetList();
      msg += ']'             ;
      cout << "(OK) The active polynomial ring was set to " << msg << '.' << endl;
      return true;
    }
    cout << "\"" << givenvalue << "\" is not a valid polynomial ring!" << endl;
    return false;
  }

  if (envvar == "stats") {
    string state = lower(getfirstword(value));
    if (state == "on") {
      Show_Stats = true;
      cout << "STATS was set to ON." << endl;
    }
    else if (state == "off") {
      Show_Stats = false;
      cout << "STATS was set to OFF." << endl;
    }
    else {
      cout << "\"STATS\" only can be set \"ON\" or \"OFF\"." << endl;
      return false;
    }
    return true;
  }

  if (envvar == "termord") {
    string to = lower(getfirstword(value)), elimnr;
    if (to == "lex") {
      termorder(Lex);
      to = "Lex";
    }
    else if (to == "deglex") {
      termorder(DegLex);
      to = "DegLex";
    }
    else if (to == "degrevlex") {
      termorder(DegRevLex);
      to = "DegRevLex";
    }
    else if (to == "elim") {
      elimnr = getbrackets(value, "(");
      if (!isnumeric(elimnr)) {
        cout << "Please give a valid number for the elimination order, e.g. SET TERMORD Elim(3)!" << endl;
        return false;
      }
      termorder(Elim, val(elimnr));
      to  = "Elim("         ;
      to += str(TermOrdElim);
      to += ')'             ;
    }
    else {
      cout << "\"" << givenvalue << "\" is not a valid term ordering" << endl;
      return false;
    }
    cout << "(OK) The active term ordering was set to " << to << '.' << endl;
    return true;
  }

  cout << "Unkown setting \"" << upper(envvar) << "\"!" << endl;
}


// startup

void startparser() {
  string inputline;

  cout = dout;
  cin  = din ;
  cout << endl << endl << endl << endl << endl;

  // initialization
  RandSeed();

  // maybe write this in a startup file some time...
  setvar("ring Q[x, y, z]"  );
  setvar("termord DegRevLex");

  cout << endl;

  do {
    inputline = readline();
    evaluate(inputline);
    cout << endl;
  } while (!Parser_Quit);
}


// display variable list

void varlist() {
  unsigned long i, l = variables.length();
  if (l == 0) {
    cout << "There are no user variables defined." << endl;
    return;
  }

  list<string> vars(l, "");
  for (i = 1; i <= l; i++)
    vars[i] = variables[i].name;
  vars.sort();
  for (i = 1; i <= l; i++) {
    cout << vars[i] << " = ";
    display(vars[i]);
  }
}


/***********************/
/* variable management */
/***********************/

variable::variable() {
}

variable::variable(const int &i) {
}

variable::variable(const string &s, const char &t, int* p) {
  name = s ;
  type = t ;
  data = p;
}

list<variable> variables;

