#ifndef z_h
#define z_h

#include <cstddef>
#include <iostream>
#include <string>


/**********/
/* ring Z */
/**********/

class Z {
  friend Z        operator- (const Z&          );
  friend Z        operator* (const Z&, const Z&);
  friend Z        operator/ (const Z&, const Z&);
  friend bool     operator==(const Z&, const Z&);
  friend bool     operator< (const Z&, const Z&);
  friend Z        abs       (const Z&          );
  friend Z        sqrt      (const Z&          );

  public:

    // constructors:
    // standard constructor:
    Z():
      size(0),
      sign(1),
      start(new unsigned char[0]) {
    };
    // general constructors:
    Z(const          int   &);
    Z(const unsigned int   &);
    Z(const          long  &);
    Z(const unsigned long  &);
    Z(const          string&);
    // copy constructor:
    Z(const Z &);

    // destructor:
    ~Z() {
      delete[] start;
    }

    // internal operators:
    Z& operator++(         );
    Z  operator++(      int);
    Z& operator--(         );
    Z  operator--(      int);
    Z  operator+=(const Z& );
    Z  operator-=(const Z& );
    Z  operator*=(const Z& );
    Z  operator/=(const Z& );
    Z  operator%=(const Z& );
    Z  operator^=(      Z  );
    Z  operator =(const Z& );

    // type conversion:
    operator            int()      ;
    operator            int() const;
    operator unsigned   int()      ;
    operator unsigned   int() const;
    operator           long()      ;
    operator           long() const;
    operator unsigned  long()      ;
    operator unsigned  long() const;
    operator          short()      ;
    operator          short() const;
    operator unsigned short()      ;
    operator unsigned short() const;
    operator string        ()      ;
    operator string        () const;

    // methods:
    string classtype();

  private:
    unsigned long size;
    short sign;
    unsigned char *start;

    // methods:
    void shorten(                    );
    void shorten(const unsigned long );
    void extend (const unsigned long&);
};


// external operators:

Z        operator+ (const Z&          );
Z        operator- (const Z&          );
Z        operator+ (const Z&, const Z&);
Z        operator- (const Z&, const Z&);
Z        operator* (const Z&, const Z&);
Z        operator/ (const Z&, const Z&);
Z        operator% (const Z&, const Z&);
Z        operator^ (const Z&, const Z&);
bool     operator==(const Z&, const Z&);
bool     operator!=(const Z&, const Z&);
bool     operator< (const Z&, const Z&);
bool     operator> (const Z&, const Z&);
bool     operator<=(const Z&, const Z&);
bool     operator>=(const Z&, const Z&);
ostream& operator<<(ostream&, const Z&);
istream& operator>>(istream&,       Z&);


// external functions:

Z    abs   (const Z&);
bool iseven(const Z&);
bool isodd (const Z&);
Z    sqrt  (const Z&);

#endif

